/*
 * Copyright (C) 2000 John David Ratliff, 
 *                    Christopher Michael O'Donnel, and
 *                    Hongwei Guan
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package datastructures;

import datatypes.Dictionary;
import datastructures.KeyTree;

public class BinarySearchTree implements Dictionary {
	private KeyTree root;

	// constructore sets the root to null
	public BinarySearchTree() {
		this.setRoot(null);
	}

	// sets the root
	private void setRoot(KeyTree root) {
		this.root = root;
	}

	// returns the root
	public KeyTree getRoot() {
		return this.root;
	}

	/* finds the Node in the tree which would be the parent of a new Tree with Comparable key
	   it recursively searches left and right using the getIndicator function */

	private KeyTree findNewParent(Comparable key) {
		return this.findNewParent(key, this.getRoot());
	}

	private KeyTree findNewParent(Comparable key, KeyTree tree) {
		int indicator = this.getIndicator(key, tree);

		if (indicator == 0) {
			// if we are equal, then we already have a node with this key
			return null;
		} else if (indicator < 0) {
			// the key should be a left child. get the left tree, if available

			if (tree.getLeftChild() != null) {
				return this.findNewParent(key, tree.getLeftChild());
			}
		} else {
			// the key should be a right child. get the right tree, if available

			if (tree.getRightChild() != null) {
				return this.findNewParent(key, tree.getRightChild());
			}
		}

		// we are not equal, and there is no child for the indicator, so we have the parent
		return tree;
	}

	/* returns the comparable value of a key with another tree
	   this determines where to add the tree: left or right */

	private int getIndicator(Comparable key, KeyTree tree) {
		return key.compareTo(tree.getKey());
	}

	/* insert takes a key and a value and inserts a new KeyTree (BST Node) with the key
	   and value. insert ignores insertions with a duplicate key */

	public void insert(Comparable key, Object value) {
		KeyTree child, parent;
		int indicator;

		if (this.getRoot() == null) {
			// BST is empty, create a new root
			this.setRoot(new KeyTree(key, value));
		} else {
			// otherwise, figure out where to put the new Node
			parent = this.findNewParent(key);
			child = new KeyTree(key, value, parent);

			if (parent == null) {
				// the parent has the child's key
				// ignore duplicate keys
				return;
			}

			indicator = this.getIndicator(key, parent);

			if (indicator < 0) {
				// make it a left child
				parent.setLeftChild(child);
 			} else {
				// make it a right child
				parent.setRightChild(child);
			}
		}
	}

	/* find the Node in the tree with the Key 'key'. 
	   it recursively searches left and right using the getIndicator function */

	private KeyTree findPosition(Comparable key) {
		if (this.getRoot() == null) {
			return null;
		}

		return this.findPosition(key, this.getRoot());
	}

	private KeyTree findPosition(Comparable key, KeyTree tree) {
		int indicator = this.getIndicator(key, tree);

		if (indicator == 0) {
			// the key equals the tree's key. we found the Node.
			return tree;
		} else if (indicator < 0) {
			// the key is left of the tree. search in the left tree, if available

			if (tree.getLeftChild() != null) {
				return this.findPosition(key, tree.getLeftChild());
			}
		} else {
			// the key is right of the tree. search in the right tree, if available

			if (tree.getRightChild() != null) {
				return this.findPosition(key, tree.getRightChild());
			}
		}

		// if we didn't find it, and we can't search left or right, then it's not in the tree
		return null;
	}

	// find the successor of a Node by going right once, then left all the way

	private KeyTree findSuccessor(KeyTree tree) {
		return this.findSuccessor(true, tree);
	}

	private KeyTree findSuccessor(boolean firstTime, KeyTree tree) {
		if (firstTime) {
			// if it is the first time, return the right child
			return (this.findSuccessor(false, tree.getRightChild()));
		} else {
			// all other attempts return the left child, if available
			// if there is no left child, we have found the successor

			if (tree.getLeftChild() != null) {
				return (this.findSuccessor(false, tree.getLeftChild()));
			} else {
				return tree;
			}
		}
	}

	/* delete the Node in the tree with the Key 'key'.
	   If the key does not exist in the tree, do nothing */

	public void delete(Comparable key) {
		KeyTree successor, tree = this.findPosition(key);

		if (tree == null) {
			// there is no Node in the tree with Key 'key'
			return;
		}

		if (tree.onlyLeftChild()) {
			// if the tree has a single child, a left child
			// then replace the parent with that left child

			if (tree.isRoot()) {
				// if we delete a root with only a left child, then that
				// child will no longer have a parent. it becomes the root
				tree.getLeftChild().setParent(null);
				this.setRoot(tree.getLeftChild());
				return;
			}

			tree.getLeftChild().setParent(tree.getParent());

			if (tree.isLeftChild()) {
				tree.getParent().setLeftChild(tree.getLeftChild());
			} else {
				tree.getParent().setRightChild(tree.getLeftChild());
			}
		} else if (tree.onlyRightChild()) {
			// if the tree has a single child, a right child
			// then replace the parent with that right child

			if (tree.isRoot()) {
				// if we delete a root with only a right child, then that
				// child will no longer have a parent. it becomes the root
				tree.getRightChild().setParent(null);
				this.setRoot(tree.getRightChild());
				return;
			}

			tree.getRightChild().setParent(tree.getParent());

			if (tree.isLeftChild()) {
				tree.getParent().setLeftChild(tree.getRightChild());
			} else {
				tree.getParent().setRightChild(tree.getRightChild());
			}
		} else if (tree.hasNoChildren()) {
			// if the tree has no children, then remove the link to the node

			if (tree.isRoot()) {
				// if we are the root, and the root has no children
				// then remove the root
				this.setRoot(null);
				return;
			}

			if (tree.isLeftChild()) {
				tree.getParent().setLeftChild(null);
			} else {
				tree.getParent().setRightChild(null);
			}
 		} else {
			// if the tree has two children, then we find the successor
			// and replace the tree with it.

			if (tree.getRightChild().onlyRightChild()) {
				// the successor is only one-down
				successor = tree.getRightChild();
				tree.setKey(successor.getKey());
				tree.setValue(successor.getValue());
				tree.setRightChild(successor.getRightChild());
				successor.getRightChild().setParent(tree);

				return;
			}

			successor = this.findSuccessor(tree);

			tree.setKey(successor.getKey());
			tree.setValue(successor.getValue());

			if (successor.onlyRightChild()) {
				// if our successor has a right child, then place
				// it as the left child of the successor's parent

				successor.getParent().setLeftChild(successor.getRightChild());
				successor.getRightChild().setParent(successor.getParent());
	  	 	} else {
				// if our successor has no children, then remove
				// it from the tree

				if (successor.isLeftChild()) {
					successor.getParent().setLeftChild(null);
				} else {
					successor.getParent().setRightChild(null);
				}
			} 
		}
 	}

	// returns the value of the Node with Key 'key'
	public Object find(Comparable key) {
		KeyTree tree;

		tree = this.findPosition(key);

		if (tree != null) {
			return tree.getValue();
		}

		return null;
	}
}

