/*
 * Copyright (C) 2000 John David Ratliff, 
 *                    Christopher Michael O'Donnel, and
 *                    Hongwei Guan
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package tools;

import datatypes.List;
import datatypes.Dictionary;
import tools.DocumentIterator;

import java.util.StringTokenizer;

public class Document {
	private List list;
	private Dictionary words;

	// The delimiter for values in the tree for multiple word occurances
	private final static String delimiter = "~";

	// use the List and Dictionary provided in the constructor
	public Document(List list, Dictionary words) {
		this.list = list;
		this.words = words;
	}

	public DocumentIterator getDocumentIterator() {
		return new DocumentIterator(this);
	}

	public DocumentIterator getDocumentIterator(int index1, int index2) {
		return new DocumentIterator(this, index1, index2);
	}

	// Tokenizes the string, and adds each word to the dictionary
	private void insertWordsInLine(String line) {
		StringTokenizer words = new StringTokenizer(line);

		while (words.hasMoreTokens()) {
			this.insertWord(words.nextToken(), line);
		}
	}

	// inserts a line into the list at position 'lineNumber'
	// adds all the words to the dictionary too...
	public void insertLine(int lineNumber, String line) {
		this.list.insert(lineNumber, line);
		this.insertWordsInLine(line);
	}

	// finds a new value for a word in the dictionary after removing
	// the String line from the value
	private String findNewValue(String line, String value) {
		StringTokenizer values = new StringTokenizer(value, Document.delimiter);
		String tempValue, newValue = "";
		boolean foundLine = false;

		while (values.hasMoreTokens()) {
			tempValue = values.nextToken();

			if (!tempValue.equals(line) || foundLine) {
				// if we aren't the 'line' or we already found
				// the first instance, add it to the new String

				newValue = newValue + Document.delimiter + tempValue;
			} else {
				if (tempValue.equals(line)) {
					// if this is the line, then allow
					// multiple instances in the new value

					foundLine = true;
				}
			}
		}

		return newValue;
	}

	// deletes all the words in the dictionary from the line 'lineNumber'
	private void deleteWordsInLine(int lineNumber) {
		String word, value, newValue, line = this.getLine(lineNumber);
		StringTokenizer words = new StringTokenizer(line);

		while (words.hasMoreTokens()) {
			word = words.nextToken();
			value = (String)this.words.find(word);

			newValue = this.findNewValue(line, value);
			this.deleteWord(word);

			if (!newValue.equals("")) {
				// if there are multiple instances of this
				// word, make sure not to lose those instances

				this.insertWord(word, newValue);
			}
		}
	}

	// deletes the line at index 'lineNumber'
	public void deleteLine(int lineNumber) {
		this.deleteWordsInLine(lineNumber);
		this.list.delete(lineNumber);
	}

	// returns the line at position 'lineNumber'
	public String getLine(int lineNumber) {
		return (String)this.list.get(lineNumber);
	}

	// returns the position of the given String line
	public int getLineIndex(String line) {
		return this.list.indexOf(line);
	}

	// returns the number of lines in the Document
	public int getLineCount() {
		return this.list.size();
	}

	// adds a word to the dictionary with key 'word' and value 'line'
	public void insertWord(String word, String line) {
		String value = (String)this.words.find(word);

		if (value == null) {
			// if this is the first instance of this word
			this.words.insert(word, line);
		} else {
			// if there are multiple instances, add it accordingly
			this.insertDuplicateWord(word, line, value);
		}
	}

	// inserts a duplicate word into the list by adding multiple strings
	// to the value of the key. the strings are positioned relative to their
	// line numbers
	public void insertDuplicateWord(String word, String line, String value) {
		String tempLine, newValue = "";
		StringTokenizer lines = new StringTokenizer(value, Document.delimiter);
		int lineNumber = this.getLineIndex(line), tempLineNumber;
		boolean foundPosition = false;

		while (lines.hasMoreTokens()) {
			tempLine = lines.nextToken();
			tempLineNumber = this.getLineIndex(tempLine);

			if (!foundPosition) {
				if (lineNumber <= tempLineNumber) {
					// if we have found the position relative
					// to the other lines, add the next instance

					newValue = newValue + Document.delimiter + line;
					foundPosition = true;
				}
			}

			// save the other instances of the word
			newValue = newValue + Document.delimiter + tempLine;
		}

		if (!foundPosition) {
			// this must be the last occurance of the word in the Document
			newValue = newValue + Document.delimiter + line;
			foundPosition = true;
		}

		// we can't change values but we can delete/add them
		this.deleteWord(word);
		this.insertWord(word, newValue);
	}

	// adds the word 'word' to the string at position 'lineNumber' at the
	// word offset 'position'
	public void insertWordInLine(String word, int lineNumber, int position) {
		int wordIndex = 0;
		String newLine = "", tempWord, line = this.getLine(lineNumber);
		StringTokenizer words = new StringTokenizer(line);
		boolean foundPosition = false, atStart = true;

		while (words.hasMoreTokens()) {
			tempWord = words.nextToken();

			if (wordIndex == position) {
				// if we found the offset, then add the word
				// to the new String here

				if (atStart) {
					newLine = word;
					atStart = false;
				} else {
					newLine = newLine + " " + word;
				}

				foundPosition = true;
			}

			// now add the words from the line

			if (atStart) {
				newLine = tempWord;
				atStart = false;
			} else {
				newLine = newLine + " " + tempWord;
			}

			wordIndex++;
		}

		if (!foundPosition) {
			// if we haven't found the position yet, it must
			// be at the very end

			if (atStart) {
				newLine = word;
			} else {
				newLine = newLine + " " + word;
			}
		}

		// no replacement -- only delete/add
		this.deleteLine(lineNumber);
		this.insertLine(lineNumber, newLine);
	}

	// deletes the word in the dictionary with Key 'word'
	public void deleteWord(String word) {
		this.words.delete(word);
	}

	// returns a String based on the line without the first occurance of
	// the word 'word'
	private String findNewLine(String word, String line) {
		String newLine = "", tempWord;
		StringTokenizer words = new StringTokenizer(line);
		boolean atStart = true, foundWord = false;

		while (words.hasMoreTokens()) {
			tempWord = words.nextToken();

			if (!word.equals(tempWord) || foundWord) {
				// if we aren't the word we're trying to delete
				// or we've already found the first instance
				// then add this word to the new string

				if (atStart) {
					newLine = tempWord;
					atStart = false;
				} else {
					newLine = newLine + " " + tempWord;
				}
			} else {
				if (word.equals(tempWord)) {
					// if we found the word, allow duplicates
					// to appear in the new string

					foundWord = true;
				}
			}
		}

		return newLine;
	}

	// delete the word 'word' from the line at position 'lineNumber'
	public void deleteWordInLine(String word, int lineNumber) {
		String newLine;
		String line = this.getLine(lineNumber);

		newLine = this.findNewLine(word, line);

		// we can't replace values, but we can delete/add them
		this.deleteLine(lineNumber);
		this.insertLine(lineNumber, newLine);
	}

	// find the index of a word 'word' in the line 'line'
	private int findWordIndex(String word, String line) {
		int index = 0;
		StringTokenizer words = new StringTokenizer(line);

		while (words.hasMoreTokens()) {
			if (words.nextToken().equals(word)) {
				// if we found it, then return the index
				return index;
			}

			index++;
		}

		return -1;
	}

	// replace the word 'oldWord' with the word 'newWord'
	public void replaceWordInLine(String oldWord, String newWord) {
		String oldLine = this.findWord(oldWord), newLine;
		int index, lineNumber;

		if (oldLine == null) {
			// oldWord doesn't exist
			return;
		}

		lineNumber = this.getLineIndex(oldLine);
		index = this.findWordIndex(oldWord, oldLine);
		newLine = this.findNewLine(oldWord, oldLine);

		this.deleteWordInLine(oldWord, lineNumber);
		this.insertWordInLine(newWord, lineNumber, index);
	}

	// find word returns the line with the first instance of the word 'word'
	public String findWord(String word) {
		String line = (String)this.words.find(word);
		StringTokenizer lines;

		if (line == null) {
			return null;
		}

		// we separate multiple instances with a delimiter
		lines = new StringTokenizer(line, Document.delimiter);

		return lines.nextToken();
	}
}

