/*
 * Copyright (C) 2000 John David Ratliff, 
 *                    Christopher Michael O'Donnel, and
 *                    Hongwei Guan
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package tools;

import datatypes.List;
import datatypes.Dictionary;
import tools.Document;

import java.util.StringTokenizer;

public class LineEditor {
	private Document document;
	private int lineNumber;
	private boolean insertMode;

	// Special strings used in the class
	private final static String CURRENT_LINE = "> ";
	private final static String COMPLETE_TEXT = "HERE IS THE COMPLETE TEXT:";

	// Command strings identify the various commands
	private final static String INSERT = "$INSERT_LINES";
	private final static String DELETE = "$DELETE_LINES";
	private final static String SETLINE = "$SET_LINE";
	private final static String DONE = "$DONE";
	private final static String DISPLAY_LINES = "$DISPLAY_LINES";
	private final static String LAST_LINE = "$LAST_LINE";
	private final static String FIND_WORD = "$FIND_WORD";
	private final static String DELETE_WORD = "$DELETE_WORD";
	private final static String INSERT_WORD = "$INSERT_WORD";
	private final static String REPLACE_WORD = "$REPLACE_WORD";

	// Error strings report the various errors
	private final static String INTEGER_FORMAT_ERROR = "INTEGER FORMAT ERROR";
	private final static String MALFORMED_RANGE = "MALFORMED RANGE";
	private final static String INTEGER_EXPECTED = "INTEGER EXPECTED";
	private final static String TWO_INTEGERS_EXPECTED = "TWO INTEGER EXPECTED";
	private final static String OUT_OF_RANGE = "OUT OF RANGE";
	private final static String ONE_INTEGER_EXPECTED = "ONE INTEGER EXPECTED";
	private final static String ZERO_OR_TWO_INTEGERS_EXPECTED = "ZERO OR TWO INTEGERS EXPECTED";
	private final static String BAD_COMMAND = "BAD COMMAND";
	private final static String NO_LINES_PRESENT = "NO LINES PRESENT";
	private final static String NO_ARGUMENTS = "COMMAND TAKES 0 ARGUMENTS";
	private final static String WORD_EXPECTED = "WORD EXPECTED";
	private final static String INVALID_LINE = "INVALID LINE ERROR";
	private final static String WORD_AND_INDEX_EXPECTED = "WORD AND INDEX EXPECTED";
	private final static String TWO_WORDS_EXPECTED = "TWO WORDS EXPECTED";

	// constructor initilizes the Document using the list and dictionary
	public LineEditor(List list, Dictionary words) {
		this.setInsertMode(false);
		this.lineNumber = 0;
		this.document = new Document(list, words);
	}

	// set the insertMode
	private void setInsertMode(boolean insertMode) {
		this.insertMode = insertMode;
	}

	// return the current status of the insertMode
	private boolean getInsertMode() {
		return this.insertMode;
	}

	// display a processing error
	private void logError(String errorMsg) {
		System.out.println(errorMsg);
	}

	// insert a string into the Document
	private void insert(String line) {
		this.document.insertLine(this.lineNumber, line);
		this.lineNumber++;
	}

	// delete the lines based on the indexes in the Tokenizer
	private void delete(StringTokenizer tokens) {
		String token;
		int loop, temp, size = this.document.getLineCount(), index[] = new int[2];

		for (loop = 0; loop < 2; loop++) {
			// grab the indexes
			token = tokens.nextToken();

			try {
				index[loop] = Integer.parseInt(token);
			} catch (NumberFormatException exception) {
				// if we didn't receive an integer, 'complain'

				this.logError(LineEditor.INTEGER_FORMAT_ERROR);
				return;
			}

			if (index[loop] < 0) {
				// if either index is < 0, complain

				this.logError(LineEditor.MALFORMED_RANGE);
				return;
			}
		}

		if (index[0] > index[1]) {
			// disallow the 1st index to be greater than the second

			this.logError(LineEditor.MALFORMED_RANGE);
			return;
		}

		// Check for OUT OF RANGE errors
		if (index[0] >= size || index[1] >= size) {
			this.logError(LineEditor.OUT_OF_RANGE);
			return;
		}

		// set current line to be next line after last line deleted
 		this.lineNumber = index[0];

		// Now we actually delete the lines...
		for (loop = index[1]; loop >= index[0]; loop--) {
			this.document.deleteLine(loop);
		}
	}

	// sets the current line to be the index contained in the Tokenizer
	private void setLine(String token) {
		int lineNumber, size = this.document.getLineCount();

		try {
			lineNumber = Integer.parseInt(token);

			if (lineNumber < 0 || lineNumber > size) {
				// check for out of range errors
				this.logError(LineEditor.OUT_OF_RANGE);
			} else {
				this.lineNumber = lineNumber;
			}
		} catch (NumberFormatException exception) {
			// if we didn't receive an integer, complain
			this.logError(LineEditor.INTEGER_EXPECTED);
		}
	}

	// insert a word at a position (both contained by the Tokenizer)
	// into the current line
	private void insertWord(StringTokenizer tokens) {
		String word, line = this.document.getLine(this.lineNumber);
		StringTokenizer words;
		int index;

		if (line == null) {
			// if we are at the dummy line
			this.logError(LineEditor.INVALID_LINE);
			return;
		}

		words = new StringTokenizer(line);
		word = tokens.nextToken();

		try {
			index = Integer.parseInt(tokens.nextToken());
		} catch (NumberFormatException exception) {
			// if we didn't receive an integer index, complain
			this.logError(LineEditor.INTEGER_FORMAT_ERROR);
			return;
		}

		if (index < 0 || index > words.countTokens()) {
			// if our index is out of range, complain
			this.logError(LineEditor.OUT_OF_RANGE);
			return;
		}

		this.document.insertWordInLine(word, lineNumber, index);
	}

	// delete word 'word' from the current line
	private void deleteWord(String word) {
		if (lineNumber == this.document.getLineCount()) {
			// cannot delete from the dummy line
			this.logError(LineEditor.INVALID_LINE);
			return;
		}

		this.document.deleteWordInLine(word, lineNumber);
	}

	// replace a word with another word (both contained in the Tokenizer)
	private void replaceWord(StringTokenizer words) {
		String oldWord, newWord;

		oldWord = words.nextToken();
		newWord = words.nextToken();

		this.document.replaceWordInLine(oldWord, newWord);
	}

	// find the line with the first occurance of the word 'word'.
	// set the lineNumber to this line
	private void findWord(String word) {
		String line = this.document.findWord(word);

		if (line != null) {
			// if we found it
			System.out.println(LineEditor.CURRENT_LINE + line);
			this.lineNumber = this.document.getLineIndex(line);
		}
	}

	// display the index of the last line
	private void lastLine() {
		if (this.document.getLineCount() == 0) {
			// if there are no lines, then there is no last line
			this.logError(LineEditor.NO_LINES_PRESENT);
		} else {
			// return the index of the last line
			System.out.println(this.document.getLineCount() - 1);
		}
	}

	// display all the lines in the Document
	private void displayLines() {
		DocumentIterator iterator = this.document.getDocumentIterator();
		int currentLine = 0;

		while (iterator.hasNext()) {
			if (currentLine == this.lineNumber) {
				// display the > prompt for the current line
				System.out.println(LineEditor.CURRENT_LINE + (String)iterator.next());
			} else {
				// indent all other lines
				System.out.println("  " + (String)iterator.next());
			}

			currentLine++;
		}
	}

	// display the lines between index1 and index2 (both in the Tokenizer)
	private void displayLines(StringTokenizer tokens) {
		int currentLine, index[] = new int[2], loop, size = this.document.getLineCount();
		String token;
		DocumentIterator iterator;

		for (loop = 0; loop < 2; loop++) {
			// grab the indexes
			token = tokens.nextToken();

			try {
				index[loop] = Integer.parseInt(token);
			} catch (NumberFormatException exception) {
				// can't use non-integer indexes, complain
				this.logError(LineEditor.INTEGER_FORMAT_ERROR);
				return;
			}

			if (index[loop] < 0 || index[loop] > (size - 1)) {
				// out of range index, complain
				this.logError(LineEditor.OUT_OF_RANGE);
				return;
			}
		}

		if (index[0] > index[1]) {
			// cannot traverse backwards
			this.logError(LineEditor.MALFORMED_RANGE);
			return;
		}

		iterator = this.document.getDocumentIterator(index[0], index[1]);
		currentLine = index[0];

		while (iterator.hasNext()) {
			if (currentLine == this.lineNumber) {
				// print the > prompt for the current line
				System.out.println(LineEditor.CURRENT_LINE + (String)iterator.next());
			} else {
				// indent all other lines
				System.out.println("  " + (String)iterator.next());
			}

			currentLine++;
		}
	}

	// print all the lines in the Document and exit the program
	private void done(StringTokenizer tokens) {
		System.out.println(LineEditor.COMPLETE_TEXT);

		this.displayLines();
		System.exit(0);
	}

	// take a line and process the input accordingly
	public void process(String line) {
		StringTokenizer tokens = new StringTokenizer(line);
		String token;
		int lower, upper;

		if (line.length() == 0) {
			// handle the empty line
			token = line;
		} else {
			// otherwise grab the first token (a potential command)
			token = tokens.nextToken();
		}

		// check for all commands one after another

		if (token.equals(LineEditor.INSERT)) {
			if (tokens.countTokens() > 0) {
				// we shouldn't be passing arguments
				logError(LineEditor.NO_ARGUMENTS);
			} else {
				// no arguments? good, insertMode = yes
				this.setInsertMode(true);
			}
		} else if (token.equals(LineEditor.DELETE)) {
			this.setInsertMode(false);

			if (tokens.countTokens() == 2) {
				// delete must have 2 arguments
				this.delete(tokens);
			} else {
				// otherwise, we complain
				this.logError(LineEditor.TWO_INTEGERS_EXPECTED);
			}
		} else if (token.equals(LineEditor.SETLINE)) {
			this.setInsertMode(false);

			if (tokens.countTokens() == 1) {
				// setline must have 1 argument
				this.setLine(tokens.nextToken());
			} else {
				// otherwise, we complain
				this.logError(LineEditor.ONE_INTEGER_EXPECTED);
			}
		} else if (token.equals(LineEditor.INSERT_WORD)) {
			this.setInsertMode(false);

			if (tokens.countTokens() == 2) {
				// insert word must have 2 arguments
				this.insertWord(tokens);
			} else {
				// otherwise, we complain
				this.logError(LineEditor.WORD_AND_INDEX_EXPECTED);
			}
		} else if (token.equals(LineEditor.DELETE_WORD)) {
			this.setInsertMode(false);

			if (tokens.countTokens() == 1) {
				// delete word must have 1 argument
				this.deleteWord(tokens.nextToken());
			} else {
				// otherwise, we complain
				this.logError(LineEditor.WORD_EXPECTED);
			}
		} else if (token.equals(LineEditor.REPLACE_WORD)) {
			this.setInsertMode(false);

			if (tokens.countTokens() == 2) {
				// replace word must have 2 arguments
				this.replaceWord(tokens);
			} else {
				// otherwise, we complain
				this.logError(LineEditor.TWO_WORDS_EXPECTED);
			}
		} else if (token.equals(LineEditor.FIND_WORD)) {
			this.setInsertMode(false);

			if (tokens.countTokens() == 1) {
				// find word must have 1 argument
				this.findWord(tokens.nextToken());
			} else {
				// otherwise, we complain
				this.logError(LineEditor.WORD_EXPECTED);
			}
		} else if (token.equals(LineEditor.LAST_LINE)) {
			this.setInsertMode(false);

			if (tokens.countTokens() == 0) {
				// last line must have 0 arguments
				this.lastLine();
			} else {
				// otherwise, we complain
				this.logError(LineEditor.NO_ARGUMENTS);
			}
		} else if (token.equals(LineEditor.DISPLAY_LINES)) {
			this.setInsertMode(false);

			if (tokens.countTokens() == 2 ) {
				// displayLine with indexes
				this.displayLines(tokens);
			} else if (tokens.countTokens() == 0) {
				// displayLines -- all lines
				this.displayLines();
			} else {
				// if not 2 or 0 arguments, complain
				this.logError(LineEditor.ZERO_OR_TWO_INTEGERS_EXPECTED);
			}
		} else if (token.equals(LineEditor.DONE)) {
			if (tokens.countTokens() != 0) {
				// done must have 0 arguments
				this.logError(LineEditor.NO_ARGUMENTS);
			} else {
				// otherwise, we complain
				this.done(tokens);
			}
		} else {
			// we didn't get a command

			if (this.getInsertMode()) {
				// if we're in insert mode, insert the line
				this.insert(line);
			} else {
				// otherwise, we complain
				this.logError(LineEditor.BAD_COMMAND);
			}
		}
	}
}
